program spectral

#define IKX ik1
#define IKY ik2
#define IKZ ik3

    use fft_spectral

    implicit none

        ! global number of colocation points along each direction
        integer, parameter :: gnx=DIMX, gny=DIMY, gnz=DIMZ

        ! padding: we can artificially append values in arrays to improve memory access on some architectures
        integer, parameter :: padkx=PPADKX, padky=PPADKY, padkz=PPADKZ

        ! global number of modes along each direction (we store only the positive modes in the direction kx).
        integer, parameter :: gnkx=gnx/2+1, gnky=gny, gnkz=gnz

        ! global dimensions of the storage array in physical and spectral space taking into account the padding
        integer, parameter :: gdx=2*(gnx/2+1 + padkx), gdy=gny+padky, gdz=gnz+padkz
        integer, parameter :: gdkx=gdx/2, gdky=gdy,  gdkz=gdz

        ! we define here the "local" number of modes, array dimensions etc. stored by a single processor.
        ! Contrary to an MPI code, there is no difference.
        integer, parameter :: nx=gnx, ny=gny, nz=gnz
        integer, parameter :: dx=gdx, dy=gdy, dz=gdz
        integer, parameter :: nkx=gnkx, nky=gnky, nkz=gnkz
        integer, parameter :: dkx=gdkx, dky=gdky, dkz=gdkz

        ! We also define for spectral space, the quantities that are stored along the first, second and third index of the array.
        ! Contrary to an MPI code, there is no difference between physical and spectral space (x,y,z) -> (kx,ky,kz)
        integer, parameter :: nk1=nkx, nk2=nky, nk3=nkz
        integer, parameter :: dk1=dkx, dk2=dky, dk3=dkz
        integer, parameter :: gnk1=gnkx, gnk2=gnky, gnk3=gnkz
        integer, parameter :: gdk1=gdkx, gdk2=gdky, gdk3=gdkz


        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1) :: svx, svy, svz, srho
        double precision, dimension(0:nx-1,0:ny-1,0:nz-1) :: wk_io

        ! we define the fft_work array
#if defined(JMFFT)
    double precision, dimension(100 + 2*(nx+ny+nz) + 512*max(nx,ny,nz)) :: fft_work
#elif defined(MATHKEISAN)
    double precision, dimension(2*(2*(nx+ny+nz)+195) + 4*256*max(nx,ny,nz)) :: fft_work
#else
    double precision, dimension(1), save :: fft_work
#endif

        integer(kind=4) :: nxread, nyread, nzread, flag, type_trunc
        double precision :: lx, ly, lz, dt, rtrunc_x, rtrunc_y, rtrunc_z, nu, xns, schm, omega2, instant
        logical(kind=4) :: de_aliasing, stratification, perturbative, linear
        character(len=200) :: file_name
        character(len=10) :: string_time

        ! wavenumbers, squared wavenumbers
        double precision, dimension(0:dkx-1) :: kx,kx2
        double precision, dimension(0:dky-1) :: ky,ky2
        double precision, dimension(0:dkz-1) :: kz,kz2
        double precision :: kxmax, kymax, kzmax, kxmax2, kymax2, kzmax2

        ! miscellaneous
        double precision, parameter :: epsilo = 1.D-100

        ! we ask for the data file name
        write(*,*)
        write(*,'(A24,TR3,I5,TR1,A1,TR1,I5,TR1,A1,TR1,I5)') ' Dimensions of the data:', nx,'x',ny,'x',nz
        write(*,*) 'Enter the name of the velocity file:'
        read(*,*) file_name

        ! we check the dimensions
        open (unit=8,file=file_name,form='unformatted')
        read(8) flag, nxread, nyread, nzread, lx, ly, lz, dt, de_aliasing, type_trunc, rtrunc_x, rtrunc_y, rtrunc_z, &
                nu, stratification, xns, schm, omega2, perturbative, linear
        read(8) instant
        close(8)

        if (flag /= 1 .or. nxread /= nx .or. nyread /= ny .or. nzread /= nz) then
            write(*,*) "ERROR: the data file has bad format. Failure."
            stop
        end if

        ! we display the time of the record
		call time2string(instant,string_time)
		write(*,*) '    file internal time:  ', instant

        ! wavenumbers
        write(*,*) 'Initialization of the wavenumbers...'
        call init_k(kx,kx2,ky,ky2,kz,kz2,kxmax,kxmax2,kymax,kymax2,kzmax,kzmax2)

        ! we initialize the FFT
        write(*,*) 'Initialization of the FFT...'
        call init_fft(svx,gnx,gny,gnz,gdx,gdy,gdz,fft_work)

        ! we read the velocity and density and switch to spectral space
        call read_array_3d(svx,1)
        call read_array_3d(svy,2)
        call read_array_3d(svz,3)
        call read_array_3d(srho,4)

        ! we switch the data into spectral space
        write(*,*) 'Switching the data to spectral space...'
        call fwd_fft(svx,fft_work)
        call fwd_fft(svy,fft_work)
        call fwd_fft(svz,fft_work)
        call fwd_fft(srho,fft_work)

        ! computing statistics
        write(*,*) 'Computing spectral statistics and writing the results...'
        call output3(svx,svy,svz,srho)

        contains


! ############################################################
!
!    Spectral statistics: WARNING: this function was never tested
!
! ############################################################

! *************************************************************************
    subroutine output3(svx,svy,svz,srho)
! *************************************************************************
    !****f* output/output3
    !  NAME
    !    output3
    !
    !  DESCRIPTION
    !    Calculate different statistics on velocity, vorticity, density in the spectral space and write them in a file.
    !    It gives densities over 3 different types of shells: planar (kz constant), cylindrical (axis kz) and spherical (k constant)
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !
    !  REMARK
    !    this subroutine has not been tested and has probably errors...
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx,svy,svz,srho

        integer, parameter :: nb_shells_plane=gnkz, nb_shells_cyl=max(gnkx,gnky), nb_shells_sphere=max(gnkx,gnky,gnkz)

        double precision :: energy_total, enstrophy_total, phi1_total, phi2_total, density_fluct_total

        double precision :: energy_mode, enstrophy_mode, phi1_mode, phi2_mode, density_fluct_mode

        double precision, dimension(0:nb_shells_plane-1) :: energy_plane, enstrophy_plane, phi1_plane, phi2_plane, &
                                                            density_fluct_plane,nb_modes_plane,volume_shell_plane,prefactor_plane

        double precision, dimension(0:nb_shells_cyl-1) :: energy_cyl, enstrophy_cyl, phi1_cyl, phi2_cyl, density_fluct_cyl, &
                                                          nb_modes_cyl, volume_shell_cyl, prefactor_cyl

        double precision, dimension(0:nb_shells_sphere-1) :: energy_sphere, enstrophy_sphere, phi1_sphere,phi2_sphere, &
                                                             density_fluct_sphere, nb_modes_sphere, volume_shell_sphere, &
                                                            prefactor_sphere

        integer :: ind_plane, ind_cyl, ind_sphere
        double precision :: total_volume, radius_plane, radius_cyl, radius_sphere
        integer :: ik1, ik2, ik3
        integer :: ikx_abs, ikx_sgn
        double precision :: pi
        integer :: io_err

        pi = acos(-1.)

        energy_total = 0.
        energy_plane = 0.
        energy_cyl = 0.
        energy_sphere = 0.

        enstrophy_total = 0.
        enstrophy_plane = 0.
        enstrophy_cyl = 0.
        enstrophy_sphere = 0.

        phi1_total = 0.
        phi1_plane = 0.
        phi1_cyl = 0.
        phi1_sphere = 0.

        phi2_total = 0.
        phi2_plane = 0.
        phi2_cyl = 0.
        phi2_sphere = 0.

        density_fluct_total = 0.
        density_fluct_plane = 0.
        density_fluct_cyl = 0.
        density_fluct_sphere = 0.

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=-nk1+1,nk1-1

                    ! we must be careful to take into account the negative kx and also count only once the modes with kx=0
                    ikx_abs = abs(IKX)
                    ikx_sgn = sign(1,IKX)

                    ! we determine the belonging planar, cylindrical and spherical shell of the current mode. The radius we keep
                    ! are in [0;1]. For a 2D simulation, there is only one plane and cylindrical/spherical shells are identical.
                    if (gnkz /= 1) then
                        radius_plane = sqrt(kz2(IKZ))/kzmax
                        radius_cyl = sqrt(kx2(ikx_abs) + ky2(IKY))/min(kxmax,kymax)
                        radius_sphere = sqrt(kx2(ikx_abs) + ky2(IKY) + kz2(IKZ))/min(kxmax,kymax,kzmax)
                    else
                        radius_plane = 0.
                        radius_cyl = sqrt(kx2(ikx_abs) + ky2(IKY))/min(kxmax,kymax)
                        radius_sphere = sqrt(kx2(ikx_abs) + ky2(IKY))/min(kxmax,kymax)
                    endif

                    ! We get the corresponding index in the results arrays. The index we will keep are in [0,nb_shells_*].
                    ind_plane    = int((nb_shells_plane-1)*radius_plane)
                    ind_cyl = int((nb_shells_cyl-1)*radius_cyl)
                    ind_sphere   = int((nb_shells_sphere-1)*radius_sphere)

                    energy_mode = 0.5*sum(svx(:,ik1,ik2,ik3)**2 + svy(:,ik1,ik2,ik3)**2 + svz(:,ik1,ik2,ik3)**2)

                    enstrophy_mode = (kx2(ikx_abs)+ky2(IKY)+kz2(IKZ))*energy_mode

                    phi1_mode = ((ky(IKY)*svx(1,ik1,ik2,ik3) - ikx_sgn*kx(ikx_abs)*svy(1,ik1,ik2,ik3))**2 + &
                                 (ky(IKY)*svx(2,ik1,ik2,ik3) - ikx_sgn*kx(ikx_abs)*svy(2,ik1,ik2,ik3))**2) / &
                                  max(kx2(ikx_abs)+ky2(IKY), epsilo)

                    phi2_mode = sum(svz(:,ik1,ik2,ik3)**2)*(kx2(ikx_abs)+ky2(IKY)+kz2(IKZ))/max(kx2(ikx_abs)+ky2(IKY),epsilo)

                    density_fluct_mode = 0.5*sum(srho(:,ik1,ik2,ik3)**2)

                    ! we add this mode to its belonging shells
                    energy_total                 = energy_total                 + energy_mode
                    enstrophy_total              = enstrophy_total              + enstrophy_mode
                    phi1_total                   = phi1_total                   + phi1_mode
                    phi2_total                   = phi2_total                   + phi2_mode
                    density_fluct_total          = density_fluct_total          + density_fluct_mode

                    if (ind_plane <= nb_shells_plane-1) then
                        nb_modes_plane(ind_plane)           = nb_modes_plane(ind_plane)         + 1

                        energy_plane(ind_plane)             = energy_plane(ind_plane)           + energy_mode
                        enstrophy_plane(ind_plane)          = enstrophy_plane(ind_plane)        + enstrophy_mode
                        phi1_plane(ind_plane)               = phi1_plane(ind_plane)             + phi1_mode
                        phi2_plane(ind_plane)               = phi2_plane(ind_plane)             + phi2_mode
                        density_fluct_plane(ind_plane)      = density_fluct_plane(ind_plane)    + density_fluct_mode
                    endif

                    if (ind_cyl <= nb_shells_cyl-1) then
                        nb_modes_cyl(ind_cyl)         = nb_modes_cyl(ind_cyl)           + 1

                        energy_cyl(ind_cyl)           = energy_cyl(ind_cyl)             + energy_mode
                        enstrophy_cyl(ind_cyl)        = enstrophy_cyl(ind_cyl)          + enstrophy_mode
                        phi1_cyl(ind_cyl)             = phi1_cyl(ind_cyl)               + phi1_mode
                        phi2_cyl(ind_cyl)             = phi2_cyl(ind_cyl)               + phi2_mode
                        density_fluct_cyl(ind_cyl)    = density_fluct_cyl(ind_cyl)      + density_fluct_mode
                    endif

                    if (ind_sphere <= nb_shells_sphere-1) then
                        nb_modes_sphere(ind_sphere)         = nb_modes_sphere(ind_sphere)           + 1

                        energy_sphere(ind_sphere)           = energy_sphere(ind_sphere)             + energy_mode
                        enstrophy_sphere(ind_sphere)        = enstrophy_sphere(ind_sphere)          + enstrophy_mode
                        phi1_sphere(ind_sphere)             = phi1_sphere(ind_sphere)               + phi1_mode
                        phi2_sphere(ind_sphere)             = phi2_sphere(ind_sphere)               + phi2_mode
                        density_fluct_sphere(ind_sphere)    = density_fluct_sphere(ind_sphere)      + density_fluct_mode
                    endif

                end do
            end do
        end do

        ! we divide by the volume of each shell in order to get volume densities
!        if (gnz /= 1) then
!            total_volume = (gnx-1)*(gny-1)*(gnz-1)/((gnx/2+1)*(gny/2)*(gnz/2))
!        else
            total_volume = (gnx-1)*(gny-1)/((gnx/2+1)*(gny/2))
!        endif

        ! we avoid any division by 0 if the number of modes in a given shell is 0
        where (nb_modes_plane==0)    nb_modes_plane=1
        where (nb_modes_cyl==0) nb_modes_cyl=1
        where (nb_modes_sphere==0)   nb_modes_sphere=1

        volume_shell_plane = nb_modes_plane/(gnx*gny*gnz)*total_volume
        volume_shell_cyl = nb_modes_cyl/(gnx*gny*gnz)*total_volume
        volume_shell_sphere = nb_modes_sphere/(gnx*gny*gnz)*total_volume


        ! we determine the prefactors that correspond to the partial volume integration on the different shells (I'm not clear...)
        if (gnz /= 1) then
            do ind_sphere=0, nb_shells_sphere-1
                prefactor_sphere(ind_sphere) = 4*pi*(1.*ind_sphere/nb_shells_sphere)**2
            end do

            do ind_cyl=0, nb_shells_cyl-1
                prefactor_cyl(ind_cyl) = 2*pi*ind_cyl/nb_shells_cyl*(gnz-1)/(gnz/2)
            end do

            prefactor_plane(:) = (gnx-1)/(gnx/2+1)*(gny-1)/(gny/2)
        else
            do ind_sphere=0, nb_shells_sphere-1
                prefactor_sphere(ind_sphere) = 2*pi*ind_sphere/nb_shells_sphere
            end do

            do ind_cyl=0, nb_shells_cyl-1
                prefactor_cyl(ind_cyl) = 2*pi*ind_cyl/nb_shells_cyl
            end do

            prefactor_plane(:) = (gnx-1)/(gnx/2+1)*(gny-1)/(gny/2)
        end if

        energy_total = energy_total/total_volume
        energy_plane = prefactor_plane*energy_plane/volume_shell_plane
        energy_cyl = prefactor_cyl*energy_cyl/volume_shell_cyl
        energy_sphere = prefactor_sphere*energy_sphere/volume_shell_sphere

        enstrophy_total = enstrophy_total/total_volume
        enstrophy_plane = prefactor_plane*enstrophy_plane/volume_shell_plane
        enstrophy_cyl = prefactor_cyl*enstrophy_cyl/volume_shell_cyl
        enstrophy_sphere = prefactor_sphere*enstrophy_sphere/volume_shell_sphere

        phi1_total = phi1_total/total_volume
        phi1_plane = prefactor_plane*phi1_plane/volume_shell_plane
        phi1_cyl = prefactor_cyl*phi1_cyl/volume_shell_cyl
        phi1_sphere = prefactor_sphere*phi1_sphere/volume_shell_sphere

        phi2_total = phi2_total/total_volume
        phi2_plane = prefactor_plane*phi2_plane/volume_shell_plane
        phi2_cyl = prefactor_cyl*phi2_cyl/volume_shell_cyl
        phi2_sphere = prefactor_sphere*phi2_sphere/volume_shell_sphere

        density_fluct_total = density_fluct_total/total_volume
        density_fluct_plane = prefactor_plane*density_fluct_plane/volume_shell_plane
        density_fluct_cyl = prefactor_cyl*density_fluct_cyl/volume_shell_cyl
        density_fluct_sphere = prefactor_sphere*density_fluct_sphere/volume_shell_sphere

        ! we check if it the first recording in the data file_name
        open(unit=76, file='spectral_statistics.out',form='unformatted',status="old",action="read",iostat=io_err)
        if (io_err == 0) then
            read(76) flag
            close(76)
        end if

        ! le fichier n'existe pas ou alors le flag n'est pas le bon: on recre le fichier de sortie
        if (io_err /= 0 .or. flag /= 1) then
                open(unit=76, file='spectral_statistics.out',form='unformatted',status="replace",action="write")
                write(76) 1, nx, ny, nz, lx, ly, lz, dt, de_aliasing, type_trunc, rtrunc_x, rtrunc_y, rtrunc_z, &
                          nu, stratification, xns, schm, omega2, perturbative, linear
                close(76)
        end if

        ! we write the recording time and the mean quadratic energy
        ! Remark: if we add the argument action="write", the recording fails with Absoft Pro Fortran 8.2 under Mac OSX
        ! whereas it works with Intel compiler on PC or on the IDRIS NEC. Strange.
        open(unit=76, file='spectral_statistics.out',form='unformatted',status="old",position="append")
        write(76) instant

        write(76) energy_total
        write(76) energy_plane
        write(76) energy_cyl
        write(76) energy_sphere

        write(76) enstrophy_total
        write(76) enstrophy_plane
        write(76) enstrophy_cyl
        write(76) enstrophy_sphere

        write(76) phi1_total
        write(76) phi1_plane
        write(76) phi1_cyl
        write(76) phi1_sphere

        write(76) phi2_total
        write(76) phi2_plane
        write(76) phi2_cyl
        write(76) phi2_sphere

        write(76) density_fluct_total
        write(76) density_fluct_plane
        write(76) density_fluct_cyl
        write(76) density_fluct_sphere

        close(76)

    end subroutine

    ! read, stride and write a 3D field (velocity, density, vorticity)
    subroutine read_array_3d(a, type_field)

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1), intent(out) :: a
        integer, intent(in) :: type_field
        double precision, dimension(19) :: trash_header
        integer :: ix, iy, iz, position_record

        if (type_field==1) then
            position_record= 33*4 + (4+8+4) + 4 + (4+8*nx*ny*nz+4)
            open (unit=8,file=file_name,access='direct',form='unformatted',recl=position_record)
            read(8,rec=1) trash_header, wk_io
            close(8)
        else
            position_record= 33*4 + (4+8+4) + (type_field-1)*(4+8*nx*ny*nz+4) + 4
            open (unit=8,file=file_name,access='direct',form='unformatted',recl=position_record)
            read(8,rec=2) wk_io
            close(8)
        end if

        do iz=0,nz-1
            do iy=0,ny-1
                do ix=0,nx-1
                    a(ix,iy,iz) = wk_io(ix,iy,iz)
                end do
            end do
        end do

    end subroutine

! *************************************************************************
    subroutine time2string(time,string_time)
! *************************************************************************
    !****f* subfunctions/time2string
    !  NAME
    !    time2string
    !
    !  DESCRIPTION
    !    Convert time to a formatted string like "t=0000.000"
    !
    !  INPUTS
    !    time: time to convert in a string
    !
    !  OUTPUT
    !    string_time: string with the desired format
    !***

        implicit none

        double precision, intent(in) :: time
        character(len=10), intent(out) :: string_time

        write(string_time,'(A2,I4.4,A1,I3.3)') 't=',int(time),'.',nint(1000*(time-int(time)))
  end subroutine


! **************************************************************************
    subroutine init_k(kx,kx2,ky,ky2,kz,kz2,kxmax,kxmax2,kymax,kymax2,kzmax,kzmax2)
! **************************************************************************
    !****f* subfunctions/init_k
    !  NAME
    !    init_k
    !
    !  DESCRIPTION
    !    Define the wavenumbers
    !
    !  OUTPUT
    !    kx,kx2,ky,ky2,kz,kz2: the wavenumbers
    !***

        implicit none

        double precision, dimension(0:dkx-1), intent(out) :: kx, kx2
        double precision, dimension(0:dky-1), intent(out) :: ky, ky2
        double precision, dimension(0:dkz-1), intent(out) :: kz, kz2
        double precision, intent(out) :: kxmax, kymax, kzmax, kxmax2, kymax2, kzmax2
        double precision :: pi
        integer :: ik

        pi = acos(-1.)

        ! Only the positive modes are stored along the x direction.
        ! The negative modes are the complex conjugates of the positive ones.
        do ik=0,gnkx-1
            kx(ik)=ik*2.*pi/lx
        end do

        do ik=0,gnky-1
            if (ik <= gnky/2) then
                ky(ik)=ik*2.*pi/ly
            else
                ky(ik)=(ik-gnky)*2.*pi/ly
            end if
        end do

        do ik=0,gnkz-1
            if (ik <= gnkz/2) then
                kz(ik)=ik*2.*pi/lz
            else
                kz(ik)=(ik-gnkz)*2.*pi/lz
            end if
        end do

        kx2=kx*kx
        ky2=ky*ky
        kz2=kz*kz

        kxmax=(gnkx-1)*2.*pi/lx
        kymax=gnky/2*2.*pi/ly
        kzmax=gnkz/2*2.*pi/lz

        kxmax2 = kxmax**2
        kymax2 = kymax**2
        kzmax2 = kzmax**2

    end subroutine

end program spectral
