!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organization : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

#include "config.h"
#include "extended_config.h"

!****************************************************************************
module global_vars
!****************************************************************************
!****h* ns3d/global_vars
!  NAME
!    global_vars
!
!  DESCRIPTION
!    Module that contains the definition of all the variables that are global to the whole program.
!***

    implicit none


    ! ***********************************************************
    !   Number of points, modes, dimensions of the arrays etc.
    ! ***********************************************************

#ifdef MPI
    ! number of processes
    integer, parameter :: nb_proc=MPI

    ! global number of colocation points along each direction
    integer, parameter :: gnx=DIMX, gny=DIMY, gnz=DIMZ

    ! padding: we can artificially append values in arrays to improve memory access on some architectures
    ! This padding is the one that will be used on the data stored by a single processor.
    integer, parameter :: padkx=PPADKX, padky=PPADKY, padkz=PPADKZ

    ! global number of modes along each direction (we store only the positive modes in the direction kx).
    integer, parameter :: gnkx=gnx/2+1, gnky=gny, gnkz=gnz

    ! global dimensions of the storage array in physical and spectral space taking into account the padding
    integer, parameter :: gdx=2*(gnx/2+1 + padkx), gdy=gny+nb_proc*padky, gdz=gnz+nb_proc*padkz
    integer, parameter :: gdkx=gdx/2, gdky=gdy,  gdkz=gdz

    ! we define here the "local" number of modes, array dimensions etc. stored by a single processor.
    ! There is a difference with the global quantities defined above because the data are distributed between the processes.
    integer, parameter :: nx=gnx, ny=gny, nz=gnz/nb_proc
    integer, parameter :: dx=gdx, dy=gdy, dz=gdz/nb_proc
    integer, parameter :: nkx=gnkx, nky=gnky/nb_proc, nkz=gnkz
    integer, parameter :: dkx=gdkx, dky=gdky/nb_proc, dkz=gdkz

    ! We also define for spectral space, the quantities that are stored along the first, second and third index of the array.
    ! There is a transposition in MPI because of the 3D FFT: (x,y,z/nb_proc) in physical space becomes (kz,kx,ky/nb_proc)
    ! in spectral space. Note that the last dimension is splitted between the nb_proc processes.
    integer, parameter :: nk1=nkz, nk2=nkx, nk3=nky
    integer, parameter :: dk1=dkz, dk2=dkx, dk3=dky
    integer, parameter :: gnk1=gnkz, gnk2=gnkx, gnk3=gnky
    integer, parameter :: gdk1=gdkz, gdk2=gdkx, gdk3=gdky


! The definitions are easier for a sequential code because the local and global quantities are the same and because there is
! no transposition of the data between physical and spectral space.
#else
    ! global number of colocation points along each direction
    integer, parameter :: gnx=DIMX, gny=DIMY, gnz=DIMZ

    ! padding: we can artificially append values in arrays to improve memory access on some architectures
    integer, parameter :: padkx=PPADKX, padky=PPADKY, padkz=PPADKZ

    ! global number of modes along each direction (we store only the positive modes in the direction kx).
    integer, parameter :: gnkx=gnx/2+1, gnky=gny, gnkz=gnz

    ! global dimensions of the storage array in physical and spectral space taking into account the padding
    integer, parameter :: gdx=2*(gnx/2+1 + padkx), gdy=gny+padky, gdz=gnz+padkz
    integer, parameter :: gdkx=gdx/2, gdky=gdy,  gdkz=gdz

    ! we define here the "local" number of modes, array dimensions etc. stored by a single processor.
    ! Contrary to an MPI code, there is no difference.
    integer, parameter :: nx=gnx, ny=gny, nz=gnz
    integer, parameter :: dx=gdx, dy=gdy, dz=gdz
    integer, parameter :: nkx=gnkx, nky=gnky, nkz=gnkz
    integer, parameter :: dkx=gdkx, dky=gdky, dkz=gdkz

    ! We also define for spectral space, the quantities that are stored along the first, second and third index of the array.
    ! Contrary to an MPI code, there is no difference between physical and spectral space (x,y,z) -> (kx,ky,kz)
    integer, parameter :: nk1=nkx, nk2=nky, nk3=nkz
    integer, parameter :: dk1=dkx, dk2=dky, dk3=dkz
    integer, parameter :: gnk1=gnkx, gnk2=gnky, gnk3=gnkz
    integer, parameter :: gdk1=gdkx, gdk2=gdky, gdk3=gdkz
#endif



    ! *************************************************
    !   Global variables read in the file 'data.in'
    ! *************************************************

    ! simulation box size
    double precision, save :: lx,ly,lz

    ! time step and initial time
    double precision, save :: dt, begin

    ! number of time steps
    integer, save :: itmax

    ! de_aliasing required for desaliasing
    logical, save :: de_aliasing
    integer, save :: type_trunc
    double precision, save :: rtrunc_x, rtrunc_y, rtrunc_z

    ! viscosity
    double precision, save :: nu

    ! stratification and rotation
    logical, save :: stratification
    double precision, save :: xns,xnsqr,omega2,schm

    ! type of simulation: in pertubation or not, linear or not
    logical, save :: perturbative, linear

    ! input of 2D base state: reading data in a file or generating the flow
    character(len=40), save :: base2D_type

    ! input of initial velocity: reading data in a file or generating the flow
    character(len=40), save :: velo_type

    ! amplitude of the white noise (0. means no white noise)
    double precision, save :: white_noise

    ! periods (number of time steps) of the different output subroutines
    integer, save :: output1_period, output2_period, output3_period


    ! *********************************
    !   Other global variables
    ! *********************************

    ! program version
    character(len=20), parameter :: program_version = "version 2.11"

    ! coordinates of the colocation points
    double precision, dimension(0:dx-1), save :: xx
    double precision, dimension(0:dy-1), save :: yy
    double precision, dimension(0:dz-1), save :: zz

    ! base state
    double precision, dimension(0:dx-1,0:dy-1), save :: vbx,vby,vbz,wbx,wby,wbz

    ! velocity and density fields
    double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), save :: vox, voy, voz, rho

    ! de_aliasing table
    logical, dimension(0:dk1-1,0:dk2-1,0:dk3-1), save :: trunctab

    ! wavenumbers, squared wavenumbers
    double precision, dimension(0:dkx-1), save :: kx,kx2
    double precision, dimension(0:dky-1), save :: ky,ky2
    double precision, dimension(0:dkz-1), save :: kz,kz2
    double precision, save :: kxmax, kymax, kzmax, kxmax2, kymax2, kzmax2

    ! current instant
    integer, save :: it

    ! miscellaneous
    double precision, parameter :: epsilo = 1.D-100


    ! *********************************
    !   MPI code
    ! *********************************

#ifdef MPI
    integer, save :: code, rank
    include "mpif.h"
#endif


    ! *********************************************
    !   Working array necessary for the 3D FFT
    ! *********************************************

#if (defined(JMFFT) && DIMZ!=1)
    double precision, dimension(100 + 2*(nx+ny+nz) + 512*max(nx,ny,nz)) :: fft_work
#elif (defined(JMFFT) && DIMZ==1)
    double precision, dimension(100 + 2*(nx+ny) + 512*max(nx,ny)) :: fft_work
#elif (defined(MATHKEISAN) && DIMZ!=1)
    double precision, dimension(2*(2*(nx+ny+nz)+195) + 4*256*max(nx,ny,nz)) :: fft_work
#elif (defined(MATHKEISAN) && DIMZ==1)
    double precision, dimension(2*(2*(nx+ny)+130) + 6*nx*ny) :: fft_work
#elif (defined(ASL) && DIMZ!=1)
    double precision, dimension(nx+2*(ny+nz) + dx*dy*dz) :: fft_work
#elif (defined(ASL) && DIMZ==1)
    double precision, dimension(nx+2*ny + dx*dy) :: fft_work
#elif (defined(ESSL) && DIMZ==1)
    double precision, dimension(100000), save :: fft_work
#elif defined(FFTW_MPI) || defined(FFTW_MPI_OPENMP) || defined(MKL_MPI)
    double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1,2), save :: fft_work
#elif defined(MATHKEISAN_MPI)
    double precision, dimension(2*dk1*dk2*dk3*2 +2*(2*gnx+gny+gnz)+4*64 + 4*max(gnx,gny,gnz)), save :: fft_work
#elif defined(JMFFT_MPI)
    double precision, dimension(2*dk1*dk2*dk3*2 +4*8*(gnx+gny+gnz)+4*100 +4*max(4+4*gnx,8*gny,8*gnz)),save :: fft_work
#else
    double precision, dimension(1), save :: fft_work
#endif

    ! *********************************************
    !   Working array necessary to write and read fields into data files
    ! *********************************************
    double precision, dimension(0:nx-1,0:ny-1,0:nz-1), save :: wk_io


    ! *********************************
    !   Time scheme variables
    ! *********************************

    ! preexponential terms
    double precision, dimension(0:dkx-1), save :: eex, eexs
    double precision, dimension(0:dky-1), save :: eey, eeys
    double precision, dimension(0:dkz-1), save :: eez, eezs

! global working arrays used mainly in the time scheme. Warning: for the Adams-Bashforth scheme, a portion of the working
! array must not be used between two time loop iterations because the non-linear terms are stored inside.
#if defined(AB2)
    double precision, dimension(dx*dy*dz,2*4), save :: work
    character(len=50), parameter :: time_scheme_name = 'Adams-Bashforth of order 2'
#elif defined(RK2)
    double precision, dimension(dx*dy*dz,2*4), save :: work
    character(len=50), parameter :: time_scheme_name = 'Runge and Kutta of order 2'
#elif defined(RK3)
    double precision, dimension(dx*dy*dz,3*4), save :: work
    character(len=50), parameter :: time_scheme_name = 'Runge and Kutta of order 3'
#elif defined(RK4)
    double precision, dimension(dx*dy*dz,3*4), save :: work
    character(len=50), parameter :: time_scheme_name = 'Runge and Kutta of order 4'
#endif

end module
