!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organixation : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

#include "config.h"
#include "extended_config.h"

!****************************************************************************
module output
!****************************************************************************
!****h* ns3d/output
!  NAME
!    output
!
!  DESCRIPTION
!    Perform all the ouput operations
!***

    use global_vars
    use fft
    use subfunctions

    implicit none

    contains

! *************************************************************************
    subroutine output_base(vbx,vby,vbz,wbx,wby,wbz)
! *************************************************************************
    !****f* output/output_base
    !  NAME
    !    output_base
    !
    !  DESCRIPTION
    !    Record in a binary file the velocity and vorticity of the base state in physical space.
    !
    !  INPUTS
    !    vbx,vby,vbz,wbx,wby,wbz: velocity and vorticity of the base fields in physical space
    !***
        double precision, dimension(0:dx-1,0:dy-1), intent(in) :: vbx,vby,vbz,wbx,wby,wbz

        ! we first write meta-data in order to help the post-treatment of the data file and avoid errors
        ! Fortran adds before and after every recording an integer that corresponds to the size of the recording
        ! expressed in bytes. We must also take into account these extra integers while post-treating the file.

        ! the exact format file is (i and j are integers corresponding to the preceding or following recording):
        ! i,1,gnx,gny,lx,ly,j,i,vbx,vby,vbz,j,i,wbx,wby,wbz,j

        open (77,file='base2D.init',form='unformatted')
        write(77) 1, gnx, gny, lx, ly

        ! we print the base state into the file
        write(77) vbx(0:nx-1,0:ny-1),vby(0:nx-1,0:ny-1),vbz(0:nx-1,0:ny-1)
        write(77) wbx(0:nx-1,0:ny-1),wby(0:nx-1,0:ny-1),wbz(0:nx-1,0:ny-1)

        close(77)

    end subroutine


! ############################################################
!
!    Basic output
!
! ############################################################
!**********************************************************************
    subroutine output1(svx,svy,svz,srho)
!*************************************************************************
    !****f* output/output1
    !  NAME
    !    output1
    !
    !  DESCRIPTION
    !    Print the elapsed time in the time stepping loop and the expected remaining time.
    !    Print the quadratic energy and the growhrate of the velocity field.
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx,svy,svz,srho
        double precision :: quad_velo, growthrate
        double precision, save :: quad_velo_old
        double precision, save :: time_start=0., time_old, time_current, time_loop
        double precision :: trash

        ! only to avoid a compiler warning "srho is not used" while keeping an identical list of arguments for output subroutine
        ! this is really a "cosmetic" change...
        trash = srho(1,0,0,0)

        ! we calculate the time already spent in the time stepping loop and the expected remaining time
        if (time_start==0) then
            call cpu_time(time_start)
            time_current = time_start
            time_old = time_start
            time_loop = 0.
        else
            call cpu_time(time_current)
            time_loop = time_current - time_old
            time_old = time_current
        end if

        ! we determine the mean quadratic velocity and the growthrate between two successive calls to the output1 subroutine
        quad_velo = sum(svx(:,0:nk1-1,0:nk2-1,0:nk3-1)**2 + svy(:,0:nk1-1,0:nk2-1,0:nk3-1)**2 + svz(:,0:nk1-1,0:nk2-1,0:nk3-1)**2)

#ifdef MPI
        call MPI_allreduce(MPI_IN_PLACE,quad_velo,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
#endif


#ifdef MPI
        if (rank==0) then
#endif
            if (quad_velo_old>epsilo) then
                growthrate=0.5*log(quad_velo/quad_velo_old)/(output1_period*dt)
            else
                growthrate=0.
            endif

            write(*,*)
            write(*,'(A4,TR2,I6,TR3,A6,TR2,F10.3)') 'it =', it, 'time =', it*dt + begin
            write(*,'(A40,TR2,I10,TR3,I10)') '   cpu time in sec (elapsed/remaining) =',int(time_current-time_start), &
                                              int(time_loop/output1_period*(itmax-it))

            write(*,*) '  mean quadratic velocity =', quad_velo
            write(*,*) '  growthrate =', growthrate
#ifdef MPI
        end if
#endif

        quad_velo_old = quad_velo

    end subroutine




! ############################################################
!
!    Data file output
!
! ############################################################
#ifdef MPI
! *************************************************************************
    subroutine output2(svx,svy,svz,srho,wkx,wky,wkz,wkrho)
! *************************************************************************
    !****f* output/output2
    !  NAME
    !    output2
    !
    !  DESCRIPTION
    !    Record in a binary file the velocity, density and vorticity fields expressed in physical space.
    !    This version uses MPI-IO (MPI Input-Output) library. The output format is exactly the same as for
    !    the sequential version (see below).
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !***
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx,svy,svz,srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx,wky,wkz,wkrho
        integer :: descriptor
        integer, dimension(MPI_STATUS_SIZE) :: mpi_status
        character(len=50) :: file_name
        character(len=10) :: sufx

        call time2string(it*dt + begin,sufx)
        file_name='velo_rho_vort.'//sufx

        call MPI_file_open(MPI_COMM_WORLD,file_name,MPI_MODE_WRONLY+MPI_MODE_CREATE,MPI_INFO_NULL,descriptor,code)

        if (rank==0) then

            ! we first write meta-data in order to help the post-treatment of the data file and avoid errors
            ! Fortran adds before and after every recording an integer that corresponds to the sixe of the recording
            ! expressed in bytes. We must also take into account these extra integers while post-treating the file.

            ! the exact format file is (i and j are integers corresponding to the preceding or following recording):

            ! i,1,gnx,gny,gnz,lx,ly,lz,de_aliasing,type_trunc,rtrunc_x,rtrunc_y,rtrunc_z,nu,stratification,
            ! xns,schm,omega2,perturbative,linear,j,i,pvx,j,i,pvy,j,i,pvz,j,i,prho,j,i,pvortx,j,i,pvorty,j,i,pvortz,j,
            ! i,vbx,vby,vbz,j,i,wbx,wby,wbz,j

            ! Contrary to a sequential code, MPI-IO does not add automatically the size of a record at the begining and the end
            ! We add it here explicitely
            call MPI_file_write_shared(descriptor,124,              1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,1,                1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,gnx,              1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,gny,              1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,gnz,              1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,lx,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,ly,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,lz,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,dt,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,de_aliasing,      1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,type_trunc,       1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,rtrunc_x,         1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,rtrunc_y,         1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,rtrunc_z,         1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,nu,               1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,stratification,   1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,xns,              1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,schm,             1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,omega2,           1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,perturbative,     1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,linear,           1, MPI_LOGICAL          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,124,              1, MPI_INTEGER          ,mpi_status,code)

            call MPI_file_write_shared(descriptor,8,                1, MPI_INTEGER          ,mpi_status,code)
            call MPI_file_write_shared(descriptor,it*dt+begin,      1, MPI_DOUBLE_PRECISION ,mpi_status,code)
            call MPI_file_write_shared(descriptor,8,                1, MPI_INTEGER          ,mpi_status,code)
        end if

        ! we switch the velocity into physical space and we print it in the file
        wkx = svx
        wky = svy
        wkz = svz
        wkrho = srho

        call bck_fft(wkx,fft_work)
        call bck_fft(wky,fft_work)
        call bck_fft(wkz,fft_work)
        if (stratification) then
            call bck_fft(wkrho,fft_work)
        end if

        call write_array(wkx,descriptor)
        call write_array(wky,descriptor)
        call write_array(wkz,descriptor)
        call write_array(wkrho,descriptor)

        ! we determine the vorticity, we switch it into physical space and we print it in the file
        call curl(svx,svy,svz,wkx,wky,wkz)

        call bck_fft(wkx,fft_work)
        call bck_fft(wky,fft_work)
        call bck_fft(wkz,fft_work)

        call write_array(wkx,descriptor)
        call write_array(wky,descriptor)
        call write_array(wkz,descriptor)

        if (rank==0) then
            call MPI_file_write_shared(descriptor,3*8*gnx*gny,1, MPI_INTEGER,mpi_status,code)
            call MPI_file_write_shared(descriptor,vbx(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,vby(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,vbz(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,3*8*gnx*gny,1, MPI_INTEGER,mpi_status,code)

            call MPI_file_write_shared(descriptor,3*8*gnx*gny,1, MPI_INTEGER,mpi_status,code)
            call MPI_file_write_shared(descriptor,wbx(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,wby(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,wbz(0:nx-1,0:ny-1),nx*ny,MPI_DOUBLE_PRECISION,mpi_status,code)
            call MPI_file_write_shared(descriptor,3*8*gnx*gny,1, MPI_INTEGER,mpi_status,code)
        end if

        call MPI_file_close(descriptor,code)

    end subroutine

! ***********************************************************
    subroutine write_array(pa,descriptor_file)
! ***********************************************************
    !****f* output/write_array
    !  NAME
    !    write_array
    !
    !  DESCRIPTION
    !    When running in MPI, this subroutine allows to write an array that is splitted between
    !    different processes in a single data file.
    !
    !  INPUTS
    !    pa: field to record in physical space
    !    descriptor_file: MPI descriptor of the file where the field should be recorded
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(in) :: pa
        integer, intent(in) :: descriptor_file
        integer, dimension(MPI_STATUS_SIZE) :: mpi_status

        integer :: ix, iy, iz

        do iz=0,nz-1
            do iy=0,ny-1
                do ix=0,nx-1
                    wk_io(ix,iy,iz) = pa(ix,iy,iz)
                end do
            end do
        end do

        if (rank==0) then
            call MPI_file_write_shared(descriptor_file,8*gnx*gny*gnz,1, MPI_INTEGER,mpi_status,code)
        end if

        call MPI_file_write_ordered(descriptor_file,wk_io,nx*ny*nz,MPI_DOUBLE_PRECISION,mpi_status,code)

        if (rank==0) then
            call MPI_file_write_shared(descriptor_file,8*gnx*gny*gnz,1, MPI_INTEGER,mpi_status,code)
        end if

    end subroutine

#else
! *************************************************************************
    subroutine output2(svx,svy,svz,srho,wkx,wky,wkz,wkrho)
! *************************************************************************
    !****f* output/output2
    !  NAME
    !    output2
    !
    !  DESCRIPTION
    !    Record in a binary file the velocity, density and vorticity fields expressed in physical space.
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !***
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx,svy,svz,srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx,wky,wkz,wkrho
        character(len=50) :: file_name
        character(len=10) :: sufx

        ! we first write meta-data in order to help the post-treatment of the data file and avoid errors
        ! Fortran adds before and after every recording an integer that corresponds to the size of the recording
        ! expressed in bytes. We must also take into account these extra integers while post-treating the file.

        ! the exact format file is (i and j are integers corresponding to the preceding or following recording):

        ! i,1,gnx,gny,gnz,lx,ly,lz,de_aliasing,type_trunc,rtrunc_x,rtrunc_y,rtrunc_z,nu,stratification,
        ! xns,schm,omega2,perturbative,linear,j,i,pvx,j,i,pvy,j,i,pvz,j,i,prho,j,i,pvortx,j,i,pvorty,j,i,pvortz,j,
        ! i,vbx,vby,vbz,j,i,wbx,wby,wbz,j

        call time2string(it*dt + begin,sufx)
        file_name='velo_rho_vort.'//sufx

        open (76,file=file_name,form='unformatted')

#ifdef NEC_LARGE_FILE
        write(76) 124
#endif

        write(76) 1, nx, ny, nz, lx, ly, lz, dt, de_aliasing, type_trunc, rtrunc_x, rtrunc_y, rtrunc_z, &
                  nu, stratification, xns, schm, omega2, perturbative, linear

#ifdef NEC_LARGE_FILE
        write(76) 124
        write(76) 8
#endif

        write(76) it*dt + begin

#ifdef NEC_LARGE_FILE
        write(76) 8
#endif
        ! we switch the velocity into physical space and we print it in the file
        wkx = svx
        wky = svy
        wkz = svz
        wkrho = srho

        call bck_fft(wkx,fft_work)
        call bck_fft(wky,fft_work)
        call bck_fft(wkz,fft_work)
        if (stratification) then
            call bck_fft(wkrho,fft_work)
        end if

        call write_array(wkx,76)
        call write_array(wky,76)
        call write_array(wkz,76)
        call write_array(wkrho,76)

        ! we determine the vorticity, we switch it into physical space and we print it in the file
        call curl(svx,svy,svz,wkx,wky,wkz)

        call bck_fft(wkx,fft_work)
        call bck_fft(wky,fft_work)
        call bck_fft(wkz,fft_work)

        call write_array(wkx,76)
        call write_array(wky,76)
        call write_array(wkz,76)

        ! we print the base state into the file
#ifdef NEC_LARGE_FILE
        write(76) 0
#endif

        write(76) vbx(0:nx-1,0:ny-1),vby(0:nx-1,0:ny-1),vbz(0:nx-1,0:ny-1)

#ifdef NEC_LARGE_FILE
        write(76) 0
        write(76) 0
#endif

        write(76) wbx(0:nx-1,0:ny-1),wby(0:nx-1,0:ny-1),wbz(0:nx-1,0:ny-1)

#ifdef NEC_LARGE_FILE
        write(76) 0
#endif

        close(76)

    end subroutine

! ***********************************************************
    subroutine write_array(pa,id_file)
! ***********************************************************
    !****f* output/write_array
    !  NAME
    !    write_array
    !
    !  DESCRIPTION
    !    Allow to write a subsection of an array into a file without using too much stack memory which
    !    can be an issue on some systems with limited stack memory.
    !    Indeed, surprisingly the following instructions are responsible for the use of an amount of
    !    stack memory corresponding to the **sum** of all the recorded arrays:
    !        write(1) aax(0:nx-1,0:ny-1,0:nz-1)
    !        write(1) aay(0:nx-1,0:ny-1,0:nz-1)
    !        (...)
    !
    !    This subroutine makes use of a single storage array that is static and so defined in data memory.
    !
    !  INPUTS
    !    pa: field to record in physical space
    !    id_file: identifier of the file where the field should be recorded
    !***

        implicit none

        double precision, dimension(0:dx-1,0:dy-1,0:dz-1),intent(in) :: pa
        integer, intent(in) :: id_file
        integer :: ix, iy, iz

        do iz=0,nz-1
            do iy=0,ny-1
                do ix=0,nx-1
                    wk_io(ix,iy,iz) = pa(ix,iy,iz)
                end do
            end do
        end do

#ifdef NEC_LARGE_FILE
        write(id_file) 0
#endif

        write(id_file) wk_io

#ifdef NEC_LARGE_FILE
        write(id_file) 0
#endif

    end subroutine
#endif



! ############################################################
!
!    Spectral statistics: WARNING: this function was never tested
!
! ############################################################

! *************************************************************************
    subroutine output3(svx,svy,svz,srho,wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
! *************************************************************************
    !****f* output/output3
    !  NAME
    !    output3
    !
    !  DESCRIPTION
    !    Calculate different statistics on velocity, vorticity, density in the spectral space and write them in a file.
    !    It gives densities over 3 different types of shells: planar (kz constant), cylindrical (axis kz) and spherical (k constant)
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields in spectral space
    !    wkx2,wky2,wkz2,wkrho2: only used as input variables with an Adams-Bashforth time scheme. Non-linear term.
    !
    !  OUTPUT
    !    wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2: the working arrays are only modified with a Runge and Kutta time scheme
    !
    !  REMARK
    !    this subroutine has not been tested and has probably errors...
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(in) :: svx,svy,svz,srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx1,wky1,wkz1,wkrho1
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: wkx2,wky2,wkz2,wkrho2

        integer, parameter :: nb_shells_plane=gnkz, nb_shells_cyl=max(gnkx,gnky), nb_shells_sphere=max(gnkx,gnky,gnkz)

        double precision :: energy_total, enstrophy_total, energy_transfer_total, phi1_total, phi2_total, density_fluct_total, &
                            phi1_transfer_total, phi2_transfer_total, density_fluct_transfer_total

        double precision :: energy_mode, enstrophy_mode, energy_transfer_mode, phi1_mode, phi2_mode, density_fluct_mode, &
                            phi1_transfer_mode, phi2_transfer_mode, density_fluct_transfer_mode

        double precision, dimension(0:nb_shells_plane-1) :: energy_plane, enstrophy_plane, energy_transfer_plane, phi1_plane, &
                                                        phi2_plane, density_fluct_plane, phi1_transfer_plane, &
                                                        phi2_transfer_plane, density_fluct_transfer_plane, nb_modes_plane, &
                                                        volume_shell_plane, prefactor_plane

        double precision, dimension(0:nb_shells_cyl-1) :: energy_cyl, enstrophy_cyl, energy_transfer_cyl, &
                                                        phi1_cyl, phi2_cyl, density_fluct_cyl, &
                                                        phi1_transfer_cyl, phi2_transfer_cyl, &
                                                        density_fluct_transfer_cyl, nb_modes_cyl, volume_shell_cyl, &
                                                        prefactor_cyl

        double precision, dimension(0:nb_shells_sphere-1) :: energy_sphere, enstrophy_sphere, energy_transfer_sphere, &
                                                             phi1_sphere,phi2_sphere,density_fluct_sphere,phi1_transfer_sphere, &
                                                             phi2_transfer_sphere, density_fluct_transfer_sphere, &
                                                             nb_modes_sphere, volume_shell_sphere, prefactor_sphere

        integer :: ind_plane, ind_cyl, ind_sphere
        double precision :: total_volume, radius_plane, radius_cyl, radius_sphere
        integer :: ik1, ik2, ik3
        integer :: ikx_abs, ikx_sgn
        double precision :: pi
        double precision :: trash

        ! only to avoid a compiler warning "srho is not used" while keeping an identical list of arguments for output subroutine
        ! this is really a "cosmetic" change...
        trash = srho(1,0,0,0)

        ! if we use a scheme of Runge and Kutta, we have to evaluate the non-linear term first. It would have been possible to call
        ! the ouput3 subroutine inside the time scheme and avoid this extra evaluation of the non-linear term but the code would
        ! have more unorganized. The extra computational cost is low because we rarely call the output3 subroutine.
        ! If we use an Adams-Bashforth scheme, the non-linear term a time t-dt is already stored in wkx2,wky2,wkz2,wkrho2
        if (time_scheme_name=='Runge and Kutta of order 2' .or. time_scheme_name=='Runge and Kutta of order 3' .or. &
            time_scheme_name=='Runge and Kutta of order 4') then
            wkx1 = svx
            wky1 = svy
            wkz1 = svz
            wkrho1 = srho

            call timer_start("non_linear_term")
            call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
            call timer_stop("non_linear_term")
        end if

        pi = acos(-1.)

        energy_total = 0.
        energy_plane = 0.
        energy_cyl = 0.
        energy_sphere = 0.

        enstrophy_total = 0.
        enstrophy_plane = 0.
        enstrophy_cyl = 0.
        enstrophy_sphere = 0.

        energy_transfer_total = 0.
        energy_transfer_plane = 0.
        energy_transfer_cyl = 0.
        energy_transfer_sphere = 0.

        phi1_total = 0.
        phi1_plane = 0.
        phi1_cyl = 0.
        phi1_sphere = 0.

        phi2_total = 0.
        phi2_plane = 0.
        phi2_cyl = 0.
        phi2_sphere = 0.

        density_fluct_total = 0.
        density_fluct_plane = 0.
        density_fluct_cyl = 0.
        density_fluct_sphere = 0.

        phi1_transfer_total = 0.
        phi1_transfer_plane = 0.
        phi1_transfer_cyl = 0.
        phi1_transfer_sphere = 0.

        phi2_transfer_total = 0.
        phi2_transfer_plane = 0.
        phi2_transfer_cyl = 0.
        phi2_transfer_sphere = 0.

        density_fluct_transfer_total = 0.
        density_fluct_transfer_plane = 0.
        density_fluct_transfer_cyl = 0.
        density_fluct_transfer_sphere = 0.

#ifndef MPI
        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=-nk1+1,nk1-1
#else
        do ik3=0,nk3-1
            do ik2=-nk2+1,nk2-1
                do ik1=0,nk1-1
#endif

                    ! we must be careful to take into account the negative kx and also count only once the modes with kx=0
                    ikx_abs = abs(IKX)
                    ikx_sgn = sign(1,IKX)

                    ! we determine the belonging planar, cylindrical and spherical shell of the current mode. The radius we keep
                    ! are in [0;1]. For a 2D simulation, there is only one plane and cylindrical/spherical shells are identical.
                    if (gnkz /= 1) then
                        radius_plane = sqrt(kz2(IKZ))/kzmax
                        radius_cyl = sqrt(kx2(ikx_abs) + ky2(IKY))/min(kxmax,kymax)
                        radius_sphere = sqrt(kx2(ikx_abs) + ky2(IKY) + kz2(IKZ))/min(kxmax,kymax,kzmax)
                    else
                        radius_plane = 0.
                        radius_cyl = sqrt(kx2(ikx_abs) + ky2(IKY))/min(kxmax,kymax)
                        radius_sphere = sqrt(kx2(ikx_abs) + ky2(IKY))/min(kxmax,kymax)
                    endif

                    ! We get the corresponding index in the results arrays. The index we will keep are in [0,nb_shells_*].
                    ind_plane    = int((nb_shells_plane-1)*radius_plane)
                    ind_cyl = int((nb_shells_cyl-1)*radius_cyl)
                    ind_sphere   = int((nb_shells_sphere-1)*radius_sphere)

                    energy_mode = 0.5*sum(svx(:,ik1,ik2,ik3)**2 + svy(:,ik1,ik2,ik3)**2 + svz(:,ik1,ik2,ik3)**2)

                    enstrophy_mode = (kx2(ikx_abs)+ky2(IKY)+kz2(IKZ))*energy_mode

                    energy_transfer_mode = wkx2(1,ik1,ik2,ik3)*svx(1,ik1,ik2,ik3) + wkx2(2,ik1,ik2,ik3)*svx(2,ik1,ik2,ik3) + &
                                           wky2(1,ik1,ik2,ik3)*svy(1,ik1,ik2,ik3) + wky2(2,ik1,ik2,ik3)*svy(2,ik1,ik2,ik3) + &
                                           wkz2(1,ik1,ik2,ik3)*svz(1,ik1,ik2,ik3) + wkz2(2,ik1,ik2,ik3)*svz(2,ik1,ik2,ik3)

                    phi1_mode = ((ky(IKY)*svx(1,ik1,ik2,ik3) - ikx_sgn*kx(ikx_abs)*svy(1,ik1,ik2,ik3))**2 + &
                                 (ky(IKY)*svx(2,ik1,ik2,ik3) - ikx_sgn*kx(ikx_abs)*svy(2,ik1,ik2,ik3))**2) / &
                                  max(kx2(ikx_abs)+ky2(IKY), epsilo)

                    phi2_mode = sum(svz(:,ik1,ik2,ik3)**2)*(kx2(ikx_abs)+ky2(IKY)+kz2(IKZ))/max(kx2(ikx_abs)+ky2(IKY),epsilo)

                    density_fluct_mode = 0.5*sum(srho(:,ik1,ik2,ik3)**2)

                    phi1_transfer_mode = ((ky(IKY)*wkx2(1,ik1,ik2,ik3)   - ikx_sgn*kx(ikx_abs)*wky2(1,ik1,ik2,ik3))* &
                                          (ky(IKY)*svx(1,ik1,ik2,ik3)    - ikx_sgn*kx(ikx_abs)*svy(1,ik1,ik2,ik3)) - &
                                          (ky(IKY)*wkx2(2,ik1,ik2,ik3)   - ikx_sgn*kx(ikx_abs)*wky2(2,ik1,ik2,ik3))* &
                                          (ky(IKY)*svx(2,ik1,ik2,ik3)    - ikx_sgn*kx(ikx_abs)*svy(2,ik1,ik2,ik3))) / &
                                          max(kx2(ikx_abs)+ky2(IKY), epsilo)

                    phi2_transfer_mode = (wkz2(1,ik1,ik2,ik3)*svz(1,ik1,ik2,ik3) - &
                                          wkz2(2,ik1,ik2,ik3)*svz(2,ik1,ik2,ik3))* &
                                         (kx2(ikx_abs)+ky2(IKY)+kz2(IKZ)) / max(kx2(ikx_abs)+ky2(IKY), epsilo)

                    density_fluct_transfer_mode = wkrho2(1,ik1,ik2,ik3)*srho(1,ik1,ik2,ik3) + &
                                                  wkrho2(2,ik1,ik2,ik3)*srho(2,ik1,ik2,ik3)

                    ! we add this mode to its belonging shells
                    energy_total                 = energy_total                 + energy_mode
                    enstrophy_total              = enstrophy_total              + enstrophy_mode
                    energy_transfer_total        = energy_transfer_total        + energy_transfer_mode
                    phi1_total                   = phi1_total                   + phi1_mode
                    phi2_total                   = phi2_total                   + phi2_mode
                    density_fluct_total          = density_fluct_total          + density_fluct_mode
                    phi1_transfer_total          = phi1_transfer_total          + phi1_transfer_mode
                    phi2_transfer_total          = phi2_transfer_total          + phi2_transfer_mode
                    density_fluct_transfer_total = density_fluct_transfer_total + density_fluct_transfer_mode

                    if (ind_plane <= nb_shells_plane-1) then
                        nb_modes_plane(ind_plane)           = nb_modes_plane(ind_plane)         + 1

                        energy_plane(ind_plane)             = energy_plane(ind_plane)           + energy_mode
                        enstrophy_plane(ind_plane)          = enstrophy_plane(ind_plane)        + enstrophy_mode
                        energy_transfer_plane(ind_plane)    = energy_transfer_plane(ind_plane)  + energy_transfer_mode
                        phi1_plane(ind_plane)               = phi1_plane(ind_plane)             + phi1_mode
                        phi2_plane(ind_plane)               = phi2_plane(ind_plane)             + phi2_mode
                        density_fluct_plane(ind_plane)      = density_fluct_plane(ind_plane)    + density_fluct_mode
                        phi1_transfer_plane(ind_plane)      = phi1_transfer_plane(ind_plane)    + phi1_transfer_mode
                        phi2_transfer_plane(ind_plane)      = phi2_transfer_plane(ind_plane)    + phi2_transfer_mode
                        density_fluct_transfer_plane(ind_plane)=density_fluct_transfer_plane(ind_plane)+density_fluct_transfer_mode
                    endif

                    if (ind_cyl <= nb_shells_cyl-1) then
                        nb_modes_cyl(ind_cyl)         = nb_modes_cyl(ind_cyl)           + 1

                        energy_cyl(ind_cyl)           = energy_cyl(ind_cyl)             + energy_mode
                        enstrophy_cyl(ind_cyl)        = enstrophy_cyl(ind_cyl)          + enstrophy_mode
                        energy_transfer_cyl(ind_cyl)  = energy_transfer_cyl(ind_cyl)    + energy_transfer_mode
                        phi1_cyl(ind_cyl)             = phi1_cyl(ind_cyl)               + phi1_mode
                        phi2_cyl(ind_cyl)             = phi2_cyl(ind_cyl)               + phi2_mode
                        density_fluct_cyl(ind_cyl)    = density_fluct_cyl(ind_cyl)      + density_fluct_mode
                        phi1_transfer_cyl(ind_cyl)    = phi1_transfer_cyl(ind_cyl)      + phi1_transfer_mode
                        phi2_transfer_cyl(ind_cyl)    = phi2_transfer_cyl(ind_cyl)      + phi2_transfer_mode
                        density_fluct_transfer_cyl(ind_cyl) = density_fluct_transfer_cyl(ind_cyl) + density_fluct_transfer_mode
                    endif

                    if (ind_sphere <= nb_shells_sphere-1) then
                        nb_modes_sphere(ind_sphere)         = nb_modes_sphere(ind_sphere)           + 1

                        energy_sphere(ind_sphere)           = energy_sphere(ind_sphere)             + energy_mode
                        enstrophy_sphere(ind_sphere)        = enstrophy_sphere(ind_sphere)          + enstrophy_mode
                        energy_transfer_sphere(ind_sphere)  = energy_transfer_sphere(ind_sphere)    + energy_transfer_mode
                        phi1_sphere(ind_sphere)             = phi1_sphere(ind_sphere)               + phi1_mode
                        phi2_sphere(ind_sphere)             = phi2_sphere(ind_sphere)               + phi2_mode
                        density_fluct_sphere(ind_sphere)    = density_fluct_sphere(ind_sphere)      + density_fluct_mode
                        phi1_transfer_sphere(ind_sphere)    = phi1_transfer_sphere(ind_sphere)      + phi1_transfer_mode
                        phi2_transfer_sphere(ind_sphere)    = phi2_transfer_sphere(ind_sphere)      + phi2_transfer_mode
                        density_fluct_transfer_sphere(ind_sphere) = density_fluct_transfer_sphere(ind_sphere) + &
                                                                    density_fluct_transfer_mode
                    endif

                end do
            end do
        end do


#ifdef MPI
        ! I use MPI_allreduce and not MPI_reduce (which is less efficient and useless here) just to be able to avoid to write
        ! specific portion of code for rank=0 (first argument MPI_IN_PLACE) and rank!=0 (first and second arguments identical).
        ! The performances are not important in this portion of code.

        call MPI_allreduce(MPI_IN_PLACE,energy_total,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,enstrophy_total,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,energy_transfer_total,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi1_total,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi2_total,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,density_fluct_total,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi1_transfer_total,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi2_transfer_total,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,density_fluct_transfer_total,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)


        call MPI_allreduce(MPI_IN_PLACE,nb_modes_sphere,nb_shells_sphere,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,energy_sphere,nb_shells_sphere,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,enstrophy_sphere,nb_shells_sphere,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,energy_transfer_sphere,nb_shells_sphere,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi1_sphere,nb_shells_sphere,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi2_sphere,nb_shells_sphere,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,density_fluct_sphere,nb_shells_sphere,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi1_transfer_sphere,nb_shells_sphere,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi2_transfer_sphere,nb_shells_sphere,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,density_fluct_transfer_sphere,nb_shells_sphere, &
                           MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)


        call MPI_allreduce(MPI_IN_PLACE,nb_modes_cyl,nb_shells_cyl,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,energy_cyl,nb_shells_cyl,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,enstrophy_cyl,nb_shells_cyl,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,energy_transfer_cyl,nb_shells_cyl,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi1_cyl,nb_shells_cyl,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi2_cyl,nb_shells_cyl,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,density_fluct_cyl,nb_shells_cyl,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi1_transfer_cyl,nb_shells_cyl,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi2_transfer_cyl,nb_shells_cyl,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,density_fluct_transfer_cyl,nb_shells_cyl, &
                           MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)


        call MPI_allreduce(MPI_IN_PLACE,nb_modes_plane,nb_shells_plane,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,energy_plane,nb_shells_plane,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,enstrophy_plane,nb_shells_plane,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,energy_transfer_plane,nb_shells_plane,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi1_plane,nb_shells_plane,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi2_plane,nb_shells_plane,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,density_fluct_plane,nb_shells_plane,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi1_transfer_plane,nb_shells_plane,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,phi2_transfer_plane,nb_shells_plane,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)
        call MPI_allreduce(MPI_IN_PLACE,density_fluct_transfer_plane,nb_shells_plane, &
                           MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,code)


        if (rank==0) then

        ! TEST
        write(*,*) 'phi2_total', phi2_total

#endif

        ! we divide by the volume of each shell in order to get volume densities
!        if (gnz /= 1) then
!            total_volume = (gnx-1)*(gny-1)*(gnz-1)/((gnx/2+1)*(gny/2)*(gnz/2))
!        else
            total_volume = (gnx-1)*(gny-1)/((gnx/2+1)*(gny/2))
!        endif

        ! we avoid any division by 0 if the number of modes in a given shell is 0
        where (nb_modes_plane==0)    nb_modes_plane=1
        where (nb_modes_cyl==0) nb_modes_cyl=1
        where (nb_modes_sphere==0)   nb_modes_sphere=1

        volume_shell_plane = nb_modes_plane/(gnx*gny*gnz)*total_volume
        volume_shell_cyl = nb_modes_cyl/(gnx*gny*gnz)*total_volume
        volume_shell_sphere = nb_modes_sphere/(gnx*gny*gnz)*total_volume


        ! we determine the prefactors that correspond to the partial volume integration on the different shells (I'm not clear...)
        if (gnz /= 1) then
            do ind_sphere=0, nb_shells_sphere-1
                prefactor_sphere(ind_sphere) = 4*pi*(1.*ind_sphere/nb_shells_sphere)**2
            end do

            do ind_cyl=0, nb_shells_cyl-1
                prefactor_cyl(ind_cyl) = 2*pi*ind_cyl/nb_shells_cyl*(gnz-1)/(gnz/2)
            end do

            prefactor_plane(:) = (gnx-1)/(gnx/2+1)*(gny-1)/(gny/2)
        else
            do ind_sphere=0, nb_shells_sphere-1
                prefactor_sphere(ind_sphere) = 2*pi*ind_sphere/nb_shells_sphere
            end do

            do ind_cyl=0, nb_shells_cyl-1
                prefactor_cyl(ind_cyl) = 2*pi*ind_cyl/nb_shells_cyl
            end do

            prefactor_plane(:) = (gnx-1)/(gnx/2+1)*(gny-1)/(gny/2)
        end if

        energy_total = energy_total/total_volume
        energy_plane = prefactor_plane*energy_plane/volume_shell_plane
        energy_cyl = prefactor_cyl*energy_cyl/volume_shell_cyl
        energy_sphere = prefactor_sphere*energy_sphere/volume_shell_sphere

        enstrophy_total = enstrophy_total/total_volume
        enstrophy_plane = prefactor_plane*enstrophy_plane/volume_shell_plane
        enstrophy_cyl = prefactor_cyl*enstrophy_cyl/volume_shell_cyl
        enstrophy_sphere = prefactor_sphere*enstrophy_sphere/volume_shell_sphere

        energy_transfer_total = energy_transfer_total/total_volume
        energy_transfer_plane = prefactor_plane*energy_transfer_plane/volume_shell_plane
        energy_transfer_cyl = prefactor_cyl*energy_transfer_cyl/volume_shell_cyl
        energy_transfer_sphere = prefactor_sphere*energy_transfer_sphere/volume_shell_sphere

        phi1_total = phi1_total/total_volume
        phi1_plane = prefactor_plane*phi1_plane/volume_shell_plane
        phi1_cyl = prefactor_cyl*phi1_cyl/volume_shell_cyl
        phi1_sphere = prefactor_sphere*phi1_sphere/volume_shell_sphere

        phi2_total = phi2_total/total_volume
        phi2_plane = prefactor_plane*phi2_plane/volume_shell_plane
        phi2_cyl = prefactor_cyl*phi2_cyl/volume_shell_cyl
        phi2_sphere = prefactor_sphere*phi2_sphere/volume_shell_sphere

        density_fluct_total = density_fluct_total/total_volume
        density_fluct_plane = prefactor_plane*density_fluct_plane/volume_shell_plane
        density_fluct_cyl = prefactor_cyl*density_fluct_cyl/volume_shell_cyl
        density_fluct_sphere = prefactor_sphere*density_fluct_sphere/volume_shell_sphere

        phi1_transfer_total = phi1_transfer_total/total_volume
        phi1_transfer_plane = prefactor_plane*phi1_transfer_plane/volume_shell_plane
        phi1_transfer_cyl = prefactor_cyl*phi1_transfer_cyl/volume_shell_cyl
        phi1_transfer_sphere = prefactor_sphere*phi1_transfer_sphere/volume_shell_sphere

        phi2_transfer_total = phi2_transfer_total/total_volume
        phi2_transfer_plane = prefactor_plane*phi2_transfer_plane/volume_shell_plane
        phi2_transfer_cyl = prefactor_cyl*phi2_transfer_cyl/volume_shell_cyl
        phi2_transfer_sphere = prefactor_sphere*phi2_transfer_sphere/volume_shell_sphere

        density_fluct_transfer_total = density_fluct_transfer_total/total_volume
        density_fluct_transfer_plane = prefactor_plane*density_fluct_transfer_plane/volume_shell_plane
        density_fluct_transfer_cyl = prefactor_cyl*density_fluct_transfer_cyl/volume_shell_cyl
        density_fluct_transfer_sphere = prefactor_sphere*density_fluct_transfer_sphere/volume_shell_sphere

        ! we write all the results in a data file
        ! first call to the subroutine: we create the file and we write the meta-data
        if (it == output3_period) then
            open(unit=76, file='spectral_statistics.out',form='unformatted',status="replace",action="write")
            write(76) 1, nx, ny, nz, lx, ly, lz, dt, de_aliasing, type_trunc, rtrunc_x, rtrunc_y, rtrunc_z, &
                     nu, stratification, xns, schm, omega2, perturbative, linear
            close(76)
        end if

        ! we write the recording time and the mean quadratic energy
        ! Remark: if we add the argument action="write", the recording fails with Absoft Pro Fortran 8.2 under Mac OSX
        ! whereas it works with Intel compiler on PC or on the IDRIS NEC. Strange.
        open(unit=76, file='spectral_statistics.out',form='unformatted',status="old",position="append")
        write(76) it*dt + begin

        write(76) energy_total
        write(76) energy_plane
        write(76) energy_cyl
        write(76) energy_sphere

        write(76) enstrophy_total
        write(76) enstrophy_plane
        write(76) enstrophy_cyl
        write(76) enstrophy_sphere

        write(76) energy_transfer_total
        write(76) energy_transfer_plane
        write(76) energy_transfer_cyl
        write(76) energy_transfer_sphere

        write(76) phi1_total
        write(76) phi1_plane
        write(76) phi1_cyl
        write(76) phi1_sphere

        write(76) phi2_total
        write(76) phi2_plane
        write(76) phi2_cyl
        write(76) phi2_sphere

        write(76) density_fluct_total
        write(76) density_fluct_plane
        write(76) density_fluct_cyl
        write(76) density_fluct_sphere

        write(76) phi1_transfer_total
        write(76) phi1_transfer_plane
        write(76) phi1_transfer_cyl
        write(76) phi1_transfer_sphere

        write(76) phi2_transfer_total
        write(76) phi2_transfer_plane
        write(76) phi2_transfer_cyl
        write(76) phi2_transfer_sphere

        write(76) density_fluct_transfer_total
        write(76) density_fluct_transfer_plane
        write(76) density_fluct_transfer_cyl
        write(76) density_fluct_transfer_sphere

        close(76)

#ifdef MPI
        end if
#endif

    end subroutine

end module
